﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using Billing.Models;
using SQLite;
using Xamarin.Essentials;
using Resource = Billing.Languages.Resource;

namespace Billing.Store
{
    public class StoreHelper
    {
        public static readonly string PersonalFolder = Environment.GetFolderPath(Environment.SpecialFolder.Personal);
        public static readonly string CacheFolder = FileSystem.CacheDirectory;

        #region Sqlite3
        private const string dbfile = "data.db3";
        private static SQLiteAsyncConnection database;
        #endregion

        private static readonly AsyncLazy<StoreHelper> Instance = new(async () =>
        {
            var instance = new StoreHelper();
            await database.CreateTablesAsync<Category, Account, Bill>();
            var count = await database.ExecuteScalarAsync<int>("SELECT COUNT(Id) FROM [Category]");
            if (count <= 0)
            {
                // init categories
                await database.InsertAllAsync(new List<Category>
                {
                    // sample categories
                    new() { Id = 1, Name = Resource.Clothing, Icon = "clothes" },
                    new() { Id = 2, Name = Resource.Food, Icon = "food" },
                    new() { Id = 4, Name = Resource.Daily, Icon = "daily" },
                    new() { Id = 5, Name = Resource.Trans, Icon = "trans" },
                    new() { Id = 6, Name = Resource.Entertainment, Icon = "face" },
                    new() { Id = 7, Name = Resource.Learn, Icon = "learn" },
                    new() { Id = 8, Name = Resource.Medical, Icon = "medical" },
                    new() { Id = 9, Name = Resource.OtherSpending, Icon = "plus" },

                    new() { Id = 10, Type = CategoryType.Income, Name = Resource.Earnings, Icon = "#brand#btc" },
                    new() { Id = 20, Type = CategoryType.Income, Name = Resource.OtherIncome, Icon = "plus" },

                    // sub-categories
                    new() { Id = 100, ParentId = 1, Name = Resource.Jewellery, Icon = "gem" },
                    new() { Id = 101, ParentId = 1, Name = Resource.Cosmetics, Icon = "makeup" },
                    new() { Id = 102, ParentId = 2, Name = Resource.Brunch, Icon = "brunch" },
                    new() { Id = 103, ParentId = 2, Name = Resource.Dinner, Icon = "dinner" },
                    new() { Id = 104, ParentId = 2, Name = Resource.Drinks, Icon = "drink" },
                    new() { Id = 105, ParentId = 2, Name = Resource.Fruit, Icon = "fruit" },
                    new() { Id = 106, ParentId = 4, Name = Resource.UtilityBill, Icon = "bill" },
                    new() { Id = 107, ParentId = 4, Name = Resource.PropertyFee, Icon = "fee" },
                    new() { Id = 108, ParentId = 4, Name = Resource.Rent, Icon = "rent" },
                    new() { Id = 109, ParentId = 4, Name = Resource.Maintenance, Icon = "maintenance" },
                    new() { Id = 110, ParentId = 5, Name = Resource.LightRail, Icon = "rail" },
                    new() { Id = 111, ParentId = 5, Name = Resource.Taxi, Icon = "taxi" },
                    new() { Id = 112, ParentId = 6, Name = Resource.Fitness, Icon = "fitness" },
                    new() { Id = 113, ParentId = 6, Name = Resource.Party, Icon = "party" },
                    new() { Id = 200, ParentId = 10, Type = CategoryType.Income, Name = Resource.Salary, Icon = "#brand#buffer" },
                    new() { Id = 201, ParentId = 10, Type = CategoryType.Income, Name = Resource.Bonus, Icon = "dollar" },
                });
            }
            return instance;
        });

        public static async Task<List<Account>> GetAccountsAsync()
        {
            var instance = await Instance;
            return await instance.GetListAsync<Account>();
        }
        public static async Task<int> SaveAccountItemAsync(Account account)
        {
            var instance = await Instance;
            return await instance.SaveItemAsync(account);
        }
        public static async Task<int> DeleteAccountItemAsync(Account account)
        {
            var instance = await Instance;
            return await instance.DeleteItemAsync(account);
        }

        public static async Task<List<Bill>> GetBillsAsync()
        {
            var instance = await Instance;
            return await instance.GetListAsync<Bill>();
        }
        public static async Task<int> SaveBillItemAsync(Bill bill)
        {
            var instance = await Instance;
            return await instance.SaveItemAsync(bill);
        }
        public static async Task<int> DeleteBillItemAsync(Bill bill)
        {
            var instance = await Instance;
            return await instance.DeleteItemAsync(bill);
        }

        public static async Task<List<Category>> GetCategoriesAsync()
        {
            var instance = await Instance;
            return await instance.GetListAsync<Category>();
        }
        public static async Task<int> SaveCategoryItemAsync(Category category)
        {
            var instance = await Instance;
            return await instance.SaveItemAsync(category);
        }
        public static async Task<int> DeleteCategoryItemAsync(Category category)
        {
            var instance = await Instance;
            return await instance.DeleteItemAsync(category);
        }

        private StoreHelper()
        {
            database = new SQLiteAsyncConnection(Path.Combine(PersonalFolder, dbfile),
                SQLiteOpenFlags.ReadWrite |
                SQLiteOpenFlags.Create |
                SQLiteOpenFlags.SharedCache);
        }

        public Task<List<T>> GetListAsync<T>(string query, params object[] args) where T : new()
        {
            try
            {
                return database.QueryAsync<T>(query, args);
            }
            catch (Exception ex)
            {
                Helper.Error("db.read", $"failed to read db, query: {string.Format(query, args)}, error: {ex.Message}");
            }
            return default;
        }

        public Task<T> GetItemAsync<T>(int id) where T : IIdItem, new()
        {
            try
            {
                var source = new TaskCompletionSource<T>();
                Task.Run(async () =>
                {
                    var list = await database.QueryAsync<T>($"SELECT * FROM [{typeof(T).Name}] WHERE [Id] = ?", id);
                    source.SetResult(list.FirstOrDefault());
                });
                return source.Task;
            }
            catch (Exception ex)
            {
                Helper.Error("db.read", $"failed to get item, table: {typeof(T)}, id: {id}, error: {ex.Message}");
            }
            return default;
        }

        #region Helper

        private Task<List<T>> GetListAsync<T>() where T : new()
        {
            try
            {
                return database.Table<T>().ToListAsync();
            }
            catch (Exception ex)
            {
                Helper.Error("db.read", $"failed to read db, error: {ex.Message}");
            }
            return default;
        }

        private Task<int> SaveItemAsync<T>(T item) where T : IIdItem
        {
            try
            {
                if (item.Id < 0)
                {
                    return database.InsertAsync(item);
                }
                else
                {
                    return database.UpdateAsync(item);
                }
            }
            catch (Exception ex)
            {
                Helper.Error("db.write", $"failed to insert/update item, table: {typeof(T)}, id: {item.Id}, item: {item}, error: {ex.Message}");
            }
            return Task.FromResult(-1);
        }

        private Task<int> DeleteItemAsync<T>(T item) where T : IIdItem
        {
            try
            {
                return database.DeleteAsync(item);
            }
            catch (Exception ex)
            {
                Helper.Error("db.delete", $"failed to delete item, table: {typeof(T)}, id: {item.Id}, item: {item}, error: {ex.Message}");
            }
            return Task.FromResult(-1);
        }

        #endregion
    }
}
